############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Chainer.py,v 1.35 2001/01/04 08:59:18 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.18
## Notes:
##   - comments in English
##
############################################################################

from Zorp import *
from Stream import Stream
from SockAddr import SockAddrInet
from Connector import Connect
import socket

class Chainer:
	"""Helper class to establish server-side connections.
	
	Chainer is used to chain to the remote server in the top session.
	In stacked sessions chainer is None, which means to chain up
	to the parent proxy.
	
	"""
	def connectServer(self, session, host, port):
		"""Connect to the targeted server.
		
		This is an interface function to be called to chain up
		to the parent proxy (or to the remote server)

		Host and port are hints, they may be ignored (depending
		on the Chainer)

		Arguments:

		  self        -- this instance
		  
		  session     -- Session object
		  
		  host	      -- host to connect to (sockaddr).
		  
		"""
		raise NotImplementedError
		
	def establishConnection(self, session, local, remote):
		"""Function to actually establish the connection.
		
		Internal function to establish a connection with the 
		given local and remote addresses. It is used by different
		chainers after finding out where to connect.
		   
		Arguments:

		  self         -- this instance
		  
		  session      -- Session object
		  
		  local	       -- bind address
		  
		  remote       -- host to connect to
		  
		Returns:
		
		  The fd of the connection to the server

		"""
		try:
			session.setServer(remote)
			if session.isServerPermitted() == Z_ACCEPT:
				conn = Connect(local, remote)
				session.server_fd = conn.blockingConnect()
				session.server_local = conn.local
				log("core.session", 3, "%s: server_fd=%d, server_addr=%s, server_zone=%s, server_local=%s" % (session.session_id, session.server_fd, session.server_address, session.server_zone, session.server_local))
				return session.server_fd
			raise DACException
		except DACException:
			pass
		except MACException:
			pass
		return -1

		

class DirectedChainer(Chainer):
	"""Connect to a predefined server.
	
	This chainer connects to a predefined server, independent
	of the original destination. It can be used to address
	a nonrouteable host from the firewall.
	   
	Attributes:
	
	  local	     -- The local data of the connection
	  
	  remote     -- The data of the server
	  
	  forge_addr -- forge client address when connecting to the server

	Usage:

	  Service("plug",
	          DirectedChainer(SockAddr('192.168.0.2', 80)),
		  PlugProxy)

	"""
	
	def __init__(self, remote, local = None, forge_addr = 0):
		"""Initializes a DirectedChainer instance.

		Sets instance attributes based on parameters.

		Arguments:

		  self       -- this instance
		
		  local	     -- Local address of the connection
		  
		  remote     -- Address of the server
		  
		  forge_addr -- forge client address when connecting to
		                the server
			
		Returns:
		
		  none

		"""
		self.local = local
		self.remote = remote
		self.forge_addr = forge_addr
	
	def connectServer(self, session, host, port):
		"""Connect to the predefined remote server.

		This function establishes connection to the
		remote server forging the client address
		if necessary.

		Arguments:

		  self         -- this instance
		
		  session      -- The data of the current session
		  
		  host	       -- not used
		  
		  port	       -- not used
			
		Returns:
		
		  The fd given by establishConnection
		"""

		if self.forge_addr:
			local = session.client_address
			local.port = 0
		else:
			local = self.local

		return self.establishConnection(session, local, self.remote)

class TransparentChainer(Chainer):
	"""Connect to the client's original destination.

	This chainer connects to the destination originally addressed
	by the client and intercepted by the firewall.
	
	Attributes:
	
	  local		      -- bind address

	  forge_addr	      -- use client address as outgoing 
	                         local address

	  forced_port	      -- if not 0 force it as destination port

	Usage:

	  You have to add a redirect rule to your packet filter
	  configuration to force connections going through your
	  firewall to be serviced locally.

	  Make sure that the Listener port is protected by a DENY
	  rule so that clients cannot connect to it directly, otherwise
	  TransparentProxy may cause an infinite loop.

	  Example:

	    Service("http", TransparentChainer(), HttpProxy)
	"""
	
	def __init__(self, local = None, forge_addr = 0, forced_port = 0):
		"""Initializes a TransparentChainer instance.
		   
		Set various attributes as passed in as parameters.
		
		Arguments:

		  self        -- this instance
		
		  local       -- local address, may be none
		  
		  forge_addr  -- if the client address shall be forged (Optional)
		  
		  forced_port -- use this port on the remote server (Optional)
		"""
		
		self.local = local
		self.forge_addr = forge_addr
		self.forced_port = forced_port
		
	def connectServer(self, session, host, port):
		"""Connect to the remote server.
		
		Finds out the original destination of the given connection
		and connects there.

		Arguments:
		
		  self         -- this instance
		  
		  session      -- The data of the current session
		  
		  host	       -- not used
		  
		  port	       -- not used
			
		Returns:
		
		  The fd given by establishConnection
		"""

		if self.forge_addr:
			local = session.client_address
			local.port = 0
		else:
			local = self.local
				
		remote = getTransparentDestination(session.client_fd)
		if self.forced_port != 0:
			remote.port = forced_port

		return self.establishConnection(session, local, remote)

class InbandChainer(Chainer):
	"""Chainer for protocols where the connection information comes from protocol logic.
	
	This chainer can be used for protocols where the target address
	is determined during the communication with the client. An
	example would be non-transparent HTTP where the target address
	is sent in the request.
	   
	Attributes:
	
	  local       -- the address of the local end of the connection
	  
	  forge_addr  -- use the client address for outgoing local address
	"""
	
	def __init__(self, local = None, forge_addr = 0):
		"""Initializes an InbandChainer instance.

		Creates an instance, sets its attributes.
		
		Arguments:

		  self       -- this instance
		  
		  local	     --	local address
		  
		  forge_addr --	forge client address
		  
		"""
		self.local = local
		self.forge_addr = forge_addr
	
	def connectServer(self, session, host, port):
		"""Called by the proxy to connect to the remote end.

		Looks up the host in the DNS and starts
		connecting there.
		
		Arguments:

		  self         -- this instance
		  
		  session      -- Session object
		  
		  host	       -- Remote hostname as determined by the protocol
		  
		  port	       -- Remote port as determinted by the protocol
		  
		Returns:
		
		  The fd given by establishConnection

		Notes:
		
		  FIXME: clean this up
		"""
		
		#message(0, "InbandChainer.connectServer()")
		try:
			ip = socket.gethostbyname(host)
			if self.forge_addr:
				local = session.client_address
				local.port = 0
			else:
				local = self.local
			return self.establishConnection(session, local, SockAddrInet(ip, port))
		except socket.error:
			return -1

class FailoverChainer(Chainer):
	"""Connects to a group of remote servers trying them sequentially.
	
	Failover chainer tries to connect several hosts until one
	of them succeeds.
		
	Attributes:
	
	  hosts		-- The list of addresses to try
	  
	  local		-- the address of the local end of the connection
	  
	  forge_addr	-- forge client address
	  
	  current_host	-- The index of the actual host in the list

	Notes:
	
	  FIXME: support priorities, shorter than system default timeouts
	"""
	
	def __init__(self, hosts, local = None, forge_addr = 0):
		"""Initializes an instance of FailoverChainer

		Sets attributes based on parameters.
		
		Arguments:
		
		  hosts			-- The list of addresses to try
		  
		  local			-- address to bind to (Optional)
		  
		  forge_addr		-- forge client address (Optional)
		"""
			
		self.local = local
		self.forge_addr = forge_addr
		self.hosts = hosts
	
	def nextHost(self):
		"""Finds the next host to try
		   
		Increments the index (current_host), zeroes it if overflowed

		Arguments:
		
		  self  -- this instance

		"""
		self.current_host = self.current_host + 1
		if self.current_host == len(self.hosts):
			self.current_host = 0

	def getHostAddress(self):
		"""Returns the actual host to try

		Returns the actual (current_host) elem of the list (hosts)

		Arguments:
		
		  self -- this instance

		"""
		return self.hosts[self.current_host]
	
	def connectServer(self, session, host, port):
		"""Called by the proxy to connect to the remote server.
		   
		Tries to connect to each host round-robin until
		one of them succeeds.

		Arguments:

		  self          -- this instance
		  
		  session	-- The data of the current session
		  
		  host		-- not used
		  
		  port		-- not used
		  
		Returns:
		
		  The fd given by establishConnection
		
		"""

		if self.forge_addr != 0:
			local = session.client_address
			local.port = 0
		else:
			local = self.local
					
		first_attempt = self.current_host
		remote = self.getHostAddress()
		res = self.establishConnection(session, local, remote)
		self.nextHost()
		while res == -1 and self.current_host != first_attempt:
			remote = self.getHostAddress()
			res = self.establishConnection(session, local, remote)
			self.nextHost()
		return res
