############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Session.py,v 1.22 2001/01/04 08:59:19 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.9
## Notes:
##   - comments in English
##
############################################################################

"""Defines session related classes and functions. 

Sessions are hierarchically stacked into each other just like proxies.
Each session has an owner session (except for the master session which
is on the top), and variables are "inherited" from owner sessions. 
(implemented using a simple getattr wrapper) This way stacked sessions
can inherit data from encapsulating proxies. (an HTTP proxy may define an
URL and a mime-type, and stacked CVP module may inspect those values)

"""

from Zorp import *
from Zone import root_zone
import os

class AbstractSession:
	"""Abstract base class for different session types (master, or stacked)

	Both MasterSession and StackedSession are derived from this class.
	"""

	def destroy(self):
		"""Destroys the session. 
		
		Parameters:
		
		  self -- this instance
		  
	        Notes:
		
		  We close filedescriptors here, in case no proxy
		  module can be started (because of policy violations,
		  or because the module cannot be found).
		
		"""
		if self.client_fd != -1:
			os.close(self.client_fd)
		if self.server_fd != -1:
			os.close(self.server_fd)


class MasterSession(AbstractSession):
	"""
	Master session class.

	Attributes:
	
	  client_fd      -- client fd
	  
	  client_stream  -- client stream
	  
	  client_address -- SockAddr instance containing client address
	  
	  client_local   -- local address (on the firewall)
	  
	  client_zone    -- zone of the client

	  server_fd      -- server fd
	  
	  server_stream  -- server stream
	  
	  server_address -- SockAddr instance containing server address
	  
	  server_local   -- local address (on the firewall)
	  
	  server_zone    -- zone of the server

	  service        -- service instance this session runs
	  
	  session_id     -- unique identifier for this session in the
	                    format: "(firewall/service:instance id/proxy)"
	  
	  instance_id    -- the instance identifier of the service
	                    (sequence number)

	  auth           -- authentication method
	  
	"""
	
	def __init__(self):
		"""Initializes a MasterSession instance.

		Arguments:
		
		  self -- this instance
		
		"""
		self.session_id = "(%s)" % firewall_name,
		
		self.client_stream = None
		self.client_fd = -1
		self.client_address = None
		self.client_local = None
		self.client_zone = None
		
		self.server_stream = None
		self.server_fd = -1
		self.server_address = None
		self.server_local = None
		self.server_zone = None
		
	        #self.auth = None
      	        #self.auth_user = ""
	
		self.service = None
		self.instance_id = 0


	def setService(self, service):
		"""Sets the service belonging to this session.

		Stores the service reference, and recalculates the session_id
		
		Arguments:

		  self -- this instance
		  
		  service  -- Service instance
		"""
		self.service = service
		self.session_id = "(%s/%s)" % (firewall_name, service.name)

	def setClient(self, fd, addr):
		"""Set client address and perform access control.
		
		Sets the client address of the given session, and performs
		access control checks.
                   
		Arguments:

		  self   -- this instance
		
		  fd     -- fd of the client
		  
		  addr   -- sockaddr of the client

		Notes:
		
		  Stores the client fd and address, looks up the
		  client zone, and performs both DAC and MAC 
		  access control. Raises an exception on failure.

		"""
		self.client_fd = fd
		self.client_address = addr
		self.client_zone = root_zone.findZone(addr)

	def isClientPermitted(self):
		log("core.session", 3, "%s: client_fd=%d, client_addr=%s, client_zone=%s, client_local=%s" % (self.session_id, self.client_fd, self.client_address, self.client_zone, self.client_local))
		if self.client_zone.isOutboundServicePermitted(self) != Z_ACCEPT:
			message(0, "%s: Outbound service %s not permitted from zone %s" % (self.session_id, self.service, self.client_zone))
			return Z_REJECT
		return Z_ACCEPT

	def setServer(self, addr):
		"""Set the server address and perform access control checks.
		
		Stores the server address of the given connection, looks
		up server zone and performs access control and raises an
		exception upon failure.

		Arguments:

		  self -- this instance
		  
		  addr -- Server address

		"""
		self.server_address = addr
		self.server_zone = root_zone.findZone(addr)
		
	def isServerPermitted(self):
		if self.server_zone.isInboundServicePermitted(self) != Z_ACCEPT:
			message(0, "%s: Inbound service %s not permitted into zone %s" % (self.session_id, self.service, self.server_zone))
			return Z_REJECT
		return Z_ACCEPT
	
	def setServiceInstance(self, instance_id):
		"""Set service instance number and recalculate session id.
		
		Sets service instance number, and makes up a unique
		identifier for this session.

		Parameters:

		  self -- this instance
		  
		  instance_id -- unique identifier of the service instance

		"""
		self.instance_id = instance_id
		self.session_id = "(%s/%s:%d)" % (firewall_name, self.name, self.instance_id)

class StackedSession(AbstractSession):
	"""Session class for subsessions.
	
	A StackedSession is a subsession, inheriting attributes from
	its parent.
	
	Attributes:
	
	  owner     -- Parent session
	  
	  chainer   -- Chainer used to chain up to parent. If none
	               simply server_fd is used.
	"""

	def __init__(self, owner, chainer = None):
                """Initializes a StackedSession instance.
		
		Arguments:

		  self        --    this instance
		  
		  owner       --    Parent session
		  
		  chainer     --    Chainer used to chain up to parent.
                """
		self.owner = owner
		self.chainer = chainer
		
	def __getattr__(self, name):
		"""Perform attribute inheritance
		
		Wrapper to return variables from parent session, if
		not overriden in this instance.

		Arguments:

		  self         --   this instance
		  
		  name         --   Name of the attribute to get.

		Returns:
		
		  The value of the given attribute.

		"""
		try:
			return self.__dict__[name]
		except KeyError:
			return getattr(self.owner, name)

	def setProxy(self, proxy):
		"""Set the proxy name used in this subsession.

		Stores a reference to the proxy class, and modifies
		the session_id to include the proxy name.
		
		Arguments:

		  self       --     this instance
		  
		  proxy      --     Proxy class, derived from Proxy

                """
		self.session_id = "(%s/%s:%d/%s)" % (firewall_name, self.name, self.instance_id, proxy)
